@@ -60,8 +60,20 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 	peer = pubkey_hashtable_lookup(&wg->peer_hashtable, in_peer.public_key);
 	if (!peer) { /* Peer doesn't exist yet. Add a new one. */
 		if (in_peer.flags & WGPEER_REMOVE_ME)
-			return -ENODEV; /* Tried to remove a non existing peer. */
-		peer = peer_rcu_get(peer_create(wg, in_peer.public_key));
+			return -ENODEV; /* Tried to remove a non-existing peer. */
+		if (in_peer.flags & WGPEER_REMOVE_PRESHARED_KEY)
+			return -EINVAL; /* Tried to remove a psk for a non-existing peer. */
+
+		down_read(&wg->static_identity.lock);
+		if (wg->static_identity.has_identity && !memcmp(in_peer.public_key, wg->static_identity.static_public, NOISE_PUBLIC_KEY_LEN)) {
+			/* We silently ignore peers that have the same public key as the device. The reason we do it silently
+			 * is that we'd like for people to be able to reuse the same set of API calls across peers. */
+			up_read(&wg->static_identity.lock);
+			goto out;
+		}
+		up_read(&wg->static_identity.lock);
+
+		peer = peer_rcu_get(peer_create(wg, in_peer.public_key, in_peer.preshared_key));
 		if (!peer)
 			return -ENOMEM;
 		if (netdev_pub(wg)->flags & IFF_UP)
@@ -74,6 +86,16 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 		goto out;
 	}
 
+	if (in_peer.flags & WGPEER_REMOVE_PRESHARED_KEY) {
+		down_write(&peer->handshake.lock);
+		memset(&peer->handshake.preshared_key, 0, NOISE_SYMMETRIC_KEY_LEN);
+		up_write(&peer->handshake.lock);
+	} else if (memcmp(zeros, in_peer.preshared_key, WG_KEY_LEN)) {
+		down_write(&peer->handshake.lock);
+		memcpy(&peer->handshake.preshared_key, in_peer.preshared_key, NOISE_SYMMETRIC_KEY_LEN);
+		up_write(&peer->handshake.lock);
+	}
+
 	if (in_peer.endpoint.addr.sa_family == AF_INET || in_peer.endpoint.addr.sa_family == AF_INET6) {
 		struct endpoint endpoint = { { { 0 } } };
 		memcpy(&endpoint, &in_peer.endpoint, sizeof(in_peer.endpoint));
@@ -146,20 +168,22 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 		noise_set_static_identity_private_key(&wg->static_identity, NULL);
 		modified_static_identity = true;
 	} else if (memcmp(zeros, in_device.private_key, WG_KEY_LEN)) {
+		u8 public_key[NOISE_PUBLIC_KEY_LEN] = { 0 };
+		struct wireguard_peer *peer;
+		/* We remove before setting, to prevent race, which means doing two 25519-genpub ops. */
+		bool unused __attribute((unused)) = curve25519_generate_public(public_key, in_device.private_key);
+		peer = pubkey_hashtable_lookup(&wg->peer_hashtable, public_key);
+		if (peer) {
+			peer_put(peer);
+			peer_remove(peer);
+		}
+
 		noise_set_static_identity_private_key(&wg->static_identity, in_device.private_key);
 		modified_static_identity = true;
 	}
 
-	if (in_device.flags & WGDEVICE_REMOVE_PRESHARED_KEY) {
-		noise_set_static_identity_preshared_key(&wg->static_identity, NULL);
-		modified_static_identity = true;
-	} else if (memcmp(zeros, in_device.preshared_key, WG_KEY_LEN)) {
-		noise_set_static_identity_preshared_key(&wg->static_identity, in_device.preshared_key);
-		modified_static_identity = true;
-	}
-
 	if (modified_static_identity)
-		cookie_checker_precompute_keys(&wg->cookie_checker, NULL);
+		cookie_checker_precompute_device_keys(&wg->cookie_checker);
 
 	for (i = 0, offset = 0, user_peer = user_device + sizeof(struct wgdevice); i < in_device.num_peers; ++i, user_peer += offset) {
 		ret = set_peer(wg, user_peer, &offset);
@@ -189,20 +213,6 @@ static inline int use_data(struct data_remaining *data, size_t size)
 	return 0;
 }
 
-static int calculate_ipmasks_size(void *ctx, struct wireguard_peer *peer, union nf_inet_addr ip, u8 cidr, int family)
-{
-	size_t *count = ctx;
-	*count += sizeof(struct wgipmask);
-	return 0;
-}
-
-static size_t calculate_peers_size(struct wireguard_device *wg)
-{
-	size_t len = peer_total_count(wg) * sizeof(struct wgpeer);
-	routing_table_walk_ips(&wg->peer_routing_table, &len, calculate_ipmasks_size);
-	return len;
-}
-
 static int populate_ipmask(void *ctx, union nf_inet_addr ip, u8 cidr, int family)
 {
 	int ret;
@@ -243,7 +253,11 @@ static int populate_peer(struct wireguard_peer *peer, void *ctx)
 	if (ret)
 		return ret;
 
+	down_read(&peer->handshake.lock);
 	memcpy(out_peer.public_key, peer->handshake.remote_static, NOISE_PUBLIC_KEY_LEN);
+	memcpy(out_peer.preshared_key, peer->handshake.preshared_key, NOISE_SYMMETRIC_KEY_LEN);
+	up_read(&peer->handshake.lock);
+
 	read_lock_bh(&peer->endpoint_lock);
 	if (peer->endpoint.addr.sa_family == AF_INET)
 		out_peer.endpoint.addr4 = peer->endpoint.addr4;
@@ -285,7 +299,8 @@ int config_get_device(struct wireguard_device *wg, void __user *user_device)
 	mutex_lock(&wg->device_update_lock);
 
 	if (!user_device) {
-		ret = calculate_peers_size(wg);
+		ret = peer_total_count(wg) * sizeof(struct wgpeer)
+		    + routing_table_count_nodes(&wg->peer_routing_table) * sizeof(struct wgipmask);
 		goto out;
 	}
 
@@ -307,7 +322,6 @@ int config_get_device(struct wireguard_device *wg, void __user *user_device)
 	if (wg->static_identity.has_identity) {
 		memcpy(out_device.private_key, wg->static_identity.static_private, WG_KEY_LEN);
 		memcpy(out_device.public_key, wg->static_identity.static_public, WG_KEY_LEN);
-		memcpy(out_device.preshared_key, wg->static_identity.preshared_key, WG_KEY_LEN);
 	}
 	up_read(&wg->static_identity.lock);
 
